/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <james@nocrew.org>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include "utils.h"
#include "neve.h"
#include "smg.h"
#include "config.h"
#include "header.h"
#include "chloe.h"
#include "http.h"
#include "file.h"
#include "cameron.h"

int get_body (const int socket, const int fd, const BrowserWindow *paula, const int expected_length);
int make_socket (const BrowserWindow *paula);
void write_to_server (const int filedes, const char *message);
int read_from_client (const int filedes, char *buffer, const int bytes_to_read);
     
char *
get_via_http (const BrowserWindow *paula)
{

  static char header[1024];
  char *request, *body, *tmp_name;
  int header_length, body_length, tmp_file, socket = 0;
  boolean still_header;

  debug (debug_chloe, "chloe: connecting to %s%s:%d.", paula->hostname, paula->path, paula->port_number);

  socket = make_socket (paula);
  debug (debug_chloe, "chloe: socket = %d", socket);
  if (socket <= 0)
    return NULL;

  request = xmalloc (strlen (paula->path) +
		     strlen (PACKAGE) + 
		     strlen (VERSION) +
		     strlen (paula->hostname) +
		     numdigit (paula->port_number) +
		     64);
  sprintf(request, "GET %s HTTP/1.0
User-Agent: %s/%s
Host: %s:%d
Accept: */*

", paula->path, PACKAGE, VERSION, paula->hostname, paula->port_number);
  debug (debug_chloe, "---request begin---\n%s---request end---", request);

  cameron_update (paula->cameron, "Sending request to %s", paula->hostname);
  write_to_server (socket, request);
  free (request);
  cameron_update (paula->cameron, "Sent request to %s; awaiting response.", paula->hostname);
  debug (debug_chloe, "chloe: request sent.");

  /* Open a temporary file to dump the HTML to.
   *
   * We do this because we often don't know the length of the file
   * we're receiving and I didn't fancy doing multiple realloc's per
   * download. */

  tmp_name = tempnam(NULL, "alyson");
  tmp_file = open (tmp_name, O_RDWR|O_CREAT|O_EXCL, 0600);
  if (tmp_file < 0)
    fubar (SYSERR, "get_via_http: couldn't create temporary file.");

  /* Read the header; looping round if necessary */

  do 
    {
      header_length =  read_from_client (socket, header, 1024);
      cameron_update (paula->cameron, "Received header.");
      still_header = parse_header (header, header_length, tmp_file, &body_length);
    } 
  while (still_header);

  /* Snarf the body into a temporary file and then into memory */

  get_body (socket, tmp_file, paula, body_length);
  if (lseek (tmp_file, 0, SEEK_SET))
    fubar (SYSERR, "get_via_http: couldn't rewind temporary file.");
  body = read_from_file (tmp_file, "temporary cache file");
  if (unlink (tmp_name))
    fubar (SYSERR, "get_via_http: couldn't delete temporary file.");

  debug (debug_chloe, "chloe: all done, reading from file; http outie.");
  return body;
}

/* *********************************************************************************** */

int
get_body (const int socket, const int fd, const BrowserWindow *paula, const int expected_length)
{
  int res, read_so_far;
  static char c[8192];

  read_so_far = 0;

  /* Read from the socket while there is available data.  */
  do
    {
      res = read_from_client (socket, c, sizeof (c));
      if (res > 0)
        {
          if (write (fd, c, res) < res)
            return -2;
        }
      read_so_far += res;
      if (expected_length != -1)
	cameron_update (paula->cameron, "Read %d of %d bytes", read_so_far, expected_length);
      else
	cameron_update (paula->cameron, "Read %d bytes", read_so_far);
      debug (debug_chloe, "chloe: read %d bytes", res);
    } while (res > 0);
  if (res < -1)
    res = -1;
  return res;
}

int
make_socket (const BrowserWindow *paula)
{
  int sock;
  struct sockaddr_in name;
  struct hostent *hostinfo;
     
  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    {
      cameron_update (paula->cameron, "Unable to create socket.");
      return -1;
    }
     
  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons (paula->port_number);

  cameron_update (paula->cameron, "Looking up %s...", paula->hostname);

  hostinfo = gethostbyname (paula->hostname);
  if (hostinfo == NULL)
    {
      cameron_update (paula->cameron, "Unknown host: %s.", paula->hostname);
      return -1;
    }

  cameron_update (paula->cameron, "Connecting to %s", paula->hostname);

/*   memcpy (&name.sin_addr, hostinfo->h_addr_list[0], hostinfo->h_length); */
  name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
  if (connect (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
    {
      cameron_update (paula->cameron, "Couldn't connect to %s.", paula->hostname);
      return -1;
    }
     
  return sock;
}

void
write_to_server (const int filedes, const char *message)
{

  int nbytes;
     
  nbytes = write (filedes, message, strlen (message) + 1);
  if (nbytes < 0)
    fubar (SYSERR, "write_to_server: write failed. ");

}

int
read_from_client (const int filedes, char *buffer, const int bytes_to_read)
{

  int nbytes;

  nbytes = read (filedes, buffer, bytes_to_read);
  if (nbytes < 0)
    fubar (SYSERR, "read_from_client: read failed. ");
  else if (nbytes == 0)
    /* End-of-file. */
    return -1;
  else
    return nbytes;

}
     
